/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2019 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*  1. Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the
*     documentation and/or other materials provided with the distribution.
*
*  3. Neither the name of the author nor the names of its contributors may
*     be used to endorse or promote products derived from this software
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  19.05.2019  mifi  First Version.
**************************************************************************/
#define __TALCPU_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdlib.h>
#include "tal.h"

#include "alt_cache.h"
#include "alt_interrupt.h"
#include "alt_clock_manager.h"
#include "alt_globaltmr.h"
#include "alt_bridge_manager.h"
#include "alt_timers.h"
#include "alt_mmu.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define ARRAY_SIZE(array)  (sizeof(array) / sizeof(array[0]))

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static uint32_t alt_pt_storage[4096] __attribute__((section(".mmu"),aligned(0x4000)));

static uint32_t dHiResPeriod   = 0;
static uint32_t dHiResPeriod16 = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  alt_pt_alloc                                                         */
/*                                                                       */
/*  In    : size, context                                                */
/*  Out   : none                                                         */
/*  Return: context                                                      */
/*************************************************************************/
static void *alt_pt_alloc (const size_t size, void *context)
{
   if (size > sizeof(alt_pt_storage))
   {
      while (1) {};
   }

   return(context);
} /* alt_pt_alloc */

/*************************************************************************/
/*  CacheInit                                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void CacheInit (void)
{
   alt_cache_system_enable();

} /* CacheInit */

/*************************************************************************/
/*  MMUInit                                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void MMUInit (void)
{
   uint32_t *ttb1 = NULL;

   extern uint8_t __DDR_segment_start__[];
   extern uint8_t __DDR_segment_size__[];
   extern uint8_t __NonCache_segment_start__[];
   extern uint8_t __NonCache_segment_size__[];

   /* Populate the page table with sections (1 MiB regions). */
   ALT_MMU_MEM_REGION_t regions[] =
   {
      /* Cacheable Memory area: DDR memory */
      {
         .va         = (void *)__DDR_segment_start__,
         .pa         = (void *)__DDR_segment_start__,
         .size       = (uint32_t)__DDR_segment_size__,
         .access     = ALT_MMU_AP_FULL_ACCESS,
         .attributes = ALT_MMU_ATTR_WBA,
         .shareable  = ALT_MMU_TTB_S_NON_SHAREABLE,
         .execute    = ALT_MMU_TTB_XN_DISABLE,
         .security   = ALT_MMU_TTB_NS_SECURE
      },

      /* Non-cacheable Memory area: 2MB */
      {
         .va         = (void *)__NonCache_segment_start__,
         .pa         = (void *)__NonCache_segment_start__,
         .size       = (uint32_t)__NonCache_segment_size__,
         .access     = ALT_MMU_AP_FULL_ACCESS,
         .attributes = ALT_MMU_ATTR_NC,
         .shareable  = ALT_MMU_TTB_S_SHAREABLE,
         .execute    = ALT_MMU_TTB_XN_DISABLE,
         .security   = ALT_MMU_TTB_NS_SECURE
      },

      /* Device area: Everything else */
      {
         .va         = (void *) 0xC0000000,
         .pa         = (void *) 0xC0000000,
         .size       = 0x40000000,
         .access     = ALT_MMU_AP_FULL_ACCESS,
         .attributes = ALT_MMU_ATTR_DEVICE_NS,
         .shareable  = ALT_MMU_TTB_S_NON_SHAREABLE,
         .execute    = ALT_MMU_TTB_XN_ENABLE,
         .security   = ALT_MMU_TTB_NS_SECURE
      },
#if 0
      /* Non-cacheable Memory area: 1 GiB */
      {
         .va         = (void *)0x40000000,
         .pa         = (void *)0x00000000,
         .size       = 0x40000000,
         .access     = ALT_MMU_AP_FULL_ACCESS,
         .attributes = ALT_MMU_ATTR_NC,
         .shareable  = ALT_MMU_TTB_S_SHAREABLE,
         .execute    = ALT_MMU_TTB_XN_DISABLE,
         .security   = ALT_MMU_TTB_NS_SECURE
      },

      /* Device area: Everything else */
      {
         .va         = (void *)0x80000000,
         .pa         = (void *)0x40000000,
         .size       = 0x40000000,
         .access     = ALT_MMU_AP_NO_ACCESS,
         .attributes = ALT_MMU_ATTR_FAULT,
         .shareable  = ALT_MMU_TTB_S_SHAREABLE,
         .execute    = ALT_MMU_TTB_XN_ENABLE,
         .security   = ALT_MMU_TTB_NS_SECURE
      },

      /* Device area: Everything else */
      {
         .va         = (void *)0xC0000000,
         .pa         = (void *)0xC0000000,
         .size       = 0x40000000,
         .access     = ALT_MMU_AP_FULL_ACCESS,
         .attributes = ALT_MMU_ATTR_DEVICE,
         .shareable  = ALT_MMU_TTB_S_SHAREABLE,
         .execute    = ALT_MMU_TTB_XN_ENABLE,
         .security   = ALT_MMU_TTB_NS_SECURE
      }
#endif
   };


   alt_mmu_init();
   alt_mmu_va_space_create(&ttb1, regions, ARRAY_SIZE(regions), alt_pt_alloc, alt_pt_storage);
   alt_mmu_va_space_enable(ttb1);

} /* MMUInit */

/*************************************************************************/
/*  SysTick_Handler                                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
//static void SysTick_Handler (void *context)
static void SysTick_Handler (uint32_t icciar, void *context)
{
   TAL_CPU_IRQ_ENTER();

   OS_TimerCallback();

   TAL_CPU_IRQ_EXIT();
} /* SysTick_Handler */

/*************************************************************************/
/*  NewSysTick_Config                                                    */
/*                                                                       */
/*  The function initializes the System Timer and its interrupt, and     */
/*  starts the System Tick Timer. Counter is in free running mode to     */
/*  generate periodic interrupts.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static uint32_t NewSysTick_Config (void)
{
   uint32_t ticks;
   uint32_t vector   = PrivateTimer_IRQn;
   uint32_t priority = SYSTICK_PRIO;

   priority &= __GIC_PRIO_MASK;
   priority  = priority << __GIC_PRIO_SHIFT;


   /* Setup Timer */
   ticks = (tal_CPUGetFrequencyPeri() / 1000);
   PRIVATE_TIMER->Load    = ticks - 1;
   PRIVATE_TIMER->Counter = ticks - 1;

   /* The timer will be started with tal_CPUSysTickStart */
//   PRIVATE_TIMER->Control = 0x7;

   /* Register interrupt function */
   tal_CPUIrqSetFunction(vector, SysTick_Handler, NULL);

   /* Set interrupt priority */
   tal_CPUIrqSetPriority(vector, SYSTICK_PRIO);

   return(ticks);
} /* NewSysTick_Config */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  tal_CPUInit                                                          */
/*                                                                       */
/*  "Initialize" the CPU.                                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUInit (void)
{
   /* Disable global CPU interrupt */
   __asm volatile ("CPSID i");

   /* Init MMU and Cache */
   MMUInit();
   CacheInit();

   /* Init the Global Interrupt Controller */
   alt_int_global_init();

   alt_int_cpu_init();
   alt_int_cpu_binary_point_set(0);

   /* Open LWH2F bridge */
   alt_bridge_init(ALT_BRIDGE_LWH2F, NULL, NULL);

   /**********************************************************/

   /* Init SysTick */
   dHiResPeriod = NewSysTick_Config();

   /*
    * dHiResPeriod value must be a 16bit count, but here it is
    * bigger. Therefore dHiResPeriod must be divided by 16.
    */
   dHiResPeriod16 = dHiResPeriod / 16;

} /* tal_CPUInit */

/*************************************************************************/
/*  tal_CPUSysTickStart                                                  */
/*                                                                       */
/*  Start the SysTick.                                                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUSysTickStart (void)
{
   /* Disable CPU interrupts */
   alt_int_cpu_disable();

   /* Enable SysTick IRQ and SysTick Timer */
   PRIVATE_TIMER->Control = 0x7;

   /* Enable systick interrupt */
   alt_int_dist_enable(PrivateTimer_IRQn);

   /* Enable global interrupts */
   alt_int_global_enable();

   /* Enable CPU interrupts */
   alt_int_cpu_enable();

} /* tal_CPUSysTickStart */

/*************************************************************************/
/*  tal_CPUIrqEnable                                                     */
/*                                                                       */
/*  Enable the given IRQ.                                                */
/*                                                                       */
/*  In    : IRQ                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqEnable (int IRQ)
{
   alt_int_dist_enable(IRQ);

} /* tal_CPUIrqEnable */

/*************************************************************************/
/*  tal_CPUIrqDisable                                                    */
/*                                                                       */
/*  Disable the given IRQ.                                               */
/*                                                                       */
/*  In    : IRQ                                                          */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqDisable (int IRQ)
{
   alt_int_dist_disable(IRQ);

} /* tal_CPUIrqDisable */

/*************************************************************************/
/*  tal_CPUIrqDisableAll                                                 */
/*                                                                       */
/*  Disable all interrupts.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqDisableAll (void)
{
   /*
    * Disable global CPU interrupt
    */
   __asm volatile ("CPSID i");

} /* tal_CPUIrqDisableAll */

/*************************************************************************/
/*  tal_CPUIrqSetPriority                                                */
/*                                                                       */
/*  Set priority of the given IRQ.                                       */
/*                                                                       */
/*  In    : IRQ, Priority                                                */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqSetPriority (int IRQ, int Priority)
{
   alt_int_dist_priority_set(IRQ, Priority);

} /* tal_CPUIrqSetPriority */

/*************************************************************************/
/*  tal_CPUIrqSetFunction                                                */
/*                                                                       */
/*  Set the handler of the given IRQ.                                    */
/*                                                                       */
/*  In    : IRQ, fnHandler                                               */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUIrqSetFunction (int IRQ, tal_irq_fnt Function, void *Context)
{
   int cpu_target = 0x1; //CPU0 will handle the interrupts

   /* Disable the interrupt first. */
   alt_int_dist_disable(IRQ);

   /* Register interrupt */
   alt_int_isr_register(IRQ, Function, Context);

   /* Set interrupt distributor target */
   alt_int_dist_target_set(IRQ, cpu_target);

} /* tal_CPUIrqSetFunction */

/*************************************************************************/
/*  tal_CPUStatGetHiResPeriod                                            */
/*                                                                       */
/*  Return the HiResPeriod value.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: HiResPeriod                                                  */
/*************************************************************************/
uint32_t tal_CPUStatGetHiResPeriod (void)
{
   return(dHiResPeriod16);
} /* tal_CPUStatGetHiResPeriod */

/*************************************************************************/
/*  tal_CPUStatGetHiResCnt                                               */
/*                                                                       */
/*  Return the HiRes counter.                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: HiRes counter                                                */
/*************************************************************************/
uint32_t tal_CPUStatGetHiResCnt (void)
{
   uint32_t dValue;

   /* Get milliseconds */
   dValue  = (OS_TimeGet() << 16);

   /* The SysTick counts down from HiResPeriod, therefore HiResPeriod - X */
   /* HiResPeriod is used, therefore divide the time by 16 too */
   dValue |= (uint16_t)((dHiResPeriod - PRIVATE_TIMER->Counter) / 16);

   return(dValue);
} /* tal_CPUStatGetHiResCnt */

/*************************************************************************/
/*  tal_CPUGetFrequencyMPU                                               */
/*                                                                       */
/*  Return the clock frequency of the MPU in MHz.                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Frequency                                                    */
/*************************************************************************/
uint32_t tal_CPUGetFrequencyMPU (void)
{
   uint32_t dFrequency;

   alt_clk_freq_get(ALT_CLK_MPU, &dFrequency);

   return(dFrequency);
} /* tal_CPUGetFrequencyMPU */

/*************************************************************************/
/*  tal_CPUGetFrequencyPeri                                              */
/*                                                                       */
/*  Return the clock frequency of the PERI bus in MHz.                   */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: Frequency                                                    */
/*************************************************************************/
uint32_t tal_CPUGetFrequencyPeri (void)
{
   uint32_t dFrequency;

   alt_clk_freq_get(ALT_CLK_MPU_PERIPH, &dFrequency);

   return(dFrequency);
} /* tal_CPUGetFrequencyPeri */

/*************************************************************************/
/*  tal_CPUDelayUS                                                       */
/*                                                                       */
/*  In    : DelayUS                                                      */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_CPUDelayUS (uint32_t DelayUS)
{
   static int         init_done = 0;
   static uint64_t    ticks_us;
   static uint32_t    timer_prescaler;
   static alt_freq_t  timer_clock;
   uint64_t           end_time;

   if (0 == init_done)
   {
      init_done = 1;

      alt_globaltmr_init();
      alt_globaltmr_autoinc_set(1);
      alt_globaltmr_start();

      timer_prescaler = alt_globaltmr_prescaler_get() + 1;
      alt_clk_freq_get(ALT_CLK_MPU_PERIPH, &timer_clock);

      ticks_us = ((timer_clock / timer_prescaler) / ALT_MICROSECS_IN_A_SEC);
   }

   end_time = alt_globaltmr_get64() + (ticks_us * DelayUS);
   while (alt_globaltmr_get64() < end_time)
   {
      __asm__ volatile ("nop");
   }

} /* tal_CPUDelayUS */

/*** EOF ***/
