/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2019 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  19.05.2019  mifi  First Version
**************************************************************************/
#define __TALCPU_COM_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include "tal.h"

#include "alt_clock_manager.h"
#include "alt_interrupt.h"
#include "soc_cv_av\socal\alt_uart.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/**
 * @brief Macro defines for UART Interrupt Enable Register
 */
#define UART_IER_RBRINT      (1 << 0)	/*!< RBR Interrupt enable */
#define UART_IER_THREINT     (1 << 1)	/*!< THR Interrupt enable */
#define UART_IER_RLSINT      (1 << 2)	/*!< RX line status interrupt enable */
#define UART_IER_MSINT       (1 << 3)	/*!< Modem status interrupt enable - valid for 11xx, 17xx/40xx UART1, 18xx/43xx UART1  only */
#define UART_IER_CTSINT      (1 << 7)	/*!< CTS signal transition interrupt enable - valid for 17xx/40xx UART1, 18xx/43xx UART1 only */
#define UART_IER_ABEOINT     (1 << 8)	/*!< Enables the end of auto-baud interrupt */
#define UART_IER_ABTOINT     (1 << 9)	/*!< Enables the auto-baud time-out interrupt */
#define UART_IER_BITMASK     (0x307)	/*!< UART interrupt enable register bit mask  - valid for 13xx, 17xx/40xx UART0/2/3, 18xx/43xx UART0/2/3 only*/
#define UART1_IER_BITMASK    (0x30F)	/*!< UART1 interrupt enable register bit mask - valid for 11xx only */
#define UART2_IER_BITMASK    (0x38F)	/*!< UART2 interrupt enable register bit mask - valid for 17xx/40xx UART1, 18xx/43xx UART1 only */

/**
 * @brief Macro defines for UART Interrupt Identification Register
 */
#define UART_IIR_INTSTAT_PEND   (1 << 0)	/*!< Interrupt pending status - Active low */
#define UART_IIR_FIFO_EN        (3 << 6)	/*!< These bits are equivalent to FCR[0] */
#define UART_IIR_ABEO_INT       (1 << 8)	/*!< End of auto-baud interrupt */
#define UART_IIR_ABTO_INT       (1 << 9)	/*!< Auto-baud time-out interrupt */
#define UART_IIR_BITMASK        (0x3CF)		/*!< UART interrupt identification register bit mask */

/* Interrupt ID bit definitions */
#define UART_IIR_INTID_MASK     (7 << 1)	/*!< Interrupt identification: Interrupt ID mask */
#define UART_IIR_INTID_RLS      (3 << 1)	/*!< Interrupt identification: Receive line interrupt */
#define UART_IIR_INTID_RDA      (2 << 1)	/*!< Interrupt identification: Receive data available interrupt */
#define UART_IIR_INTID_CTI      (6 << 1)	/*!< Interrupt identification: Character time-out indicator interrupt */
#define UART_IIR_INTID_THRE     (1 << 1)	/*!< Interrupt identification: THRE interrupt */
#define UART_IIR_INTID_MODEM    (0 << 1)	/*!< Interrupt identification: Modem interrupt */

/**
 * @brief Macro defines for UART FIFO Control Register
 */
#define UART_FCR_FIFO_EN        (1 << 0)	/*!< UART FIFO enable */
#define UART_FCR_RX_RS          (1 << 1)	/*!< UART RX FIFO reset */
#define UART_FCR_TX_RS          (1 << 2)	/*!< UART TX FIFO reset */
#define UART_FCR_DMAMODE_SEL    (1 << 3)	/*!< UART DMA mode selection - valid for 17xx/40xx, 18xx/43xx only */
#define UART_FCR_BITMASK        (0xCF)		/*!< UART FIFO control bit mask */

#define UART_TX_FIFO_SIZE       (16)

/* FIFO trigger level bit definitions */
#define UART_FCR_TRG_LEV0       (0)			/*!< UART FIFO trigger level 0: 1 character */
#define UART_FCR_TRG_LEV1       (1 << 6)	/*!< UART FIFO trigger level 1: 4 character */
#define UART_FCR_TRG_LEV2       (2 << 6)	/*!< UART FIFO trigger level 2: 8 character */
#define UART_FCR_TRG_LEV3       (3 << 6)	/*!< UART FIFO trigger level 3: 14 character */

/**
 * @brief Macro defines for UART Line Control Register
 */
/* UART word length select bit definitions */
#define UART_LCR_WLEN_MASK      (3 << 0)		/*!< UART word length select bit mask */
#define UART_LCR_WLEN5          (0 << 0)		/*!< UART word length select: 5 bit data mode */
#define UART_LCR_WLEN6          (1 << 0)		/*!< UART word length select: 6 bit data mode */
#define UART_LCR_WLEN7          (2 << 0)		/*!< UART word length select: 7 bit data mode */
#define UART_LCR_WLEN8          (3 << 0)		/*!< UART word length select: 8 bit data mode */

/* UART Stop bit select bit definitions */
#define UART_LCR_SBS_MASK       (1 << 2)		/*!< UART stop bit select: bit mask */
#define UART_LCR_SBS_1BIT       (0 << 2)		/*!< UART stop bit select: 1 stop bit */
#define UART_LCR_SBS_2BIT       (1 << 2)		/*!< UART stop bit select: 2 stop bits (in 5 bit data mode, 1.5 stop bits) */

/* UART Parity enable bit definitions */
#define UART_LCR_PARITY_EN      (1 << 3)		/*!< UART Parity Enable */
#define UART_LCR_PARITY_DIS     (0 << 3)		/*!< UART Parity Disable */
#define UART_LCR_PARITY_ODD     (0 << 4)		/*!< UART Parity select: Odd parity */
#define UART_LCR_PARITY_EVEN    (1 << 4)		/*!< UART Parity select: Even parity */
#define UART_LCR_PARITY_F_1     (2 << 4)		/*!< UART Parity select: Forced 1 stick parity */
#define UART_LCR_PARITY_F_0     (3 << 4)		/*!< UART Parity select: Forced 0 stick parity */
#define UART_LCR_BREAK_EN       (1 << 6)		/*!< UART Break transmission enable */
#define UART_LCR_DLAB_EN        (1 << 7)		/*!< UART Divisor Latches Access bit enable */
#define UART_LCR_BITMASK        (0xFF)			/*!< UART line control bit mask */

/**
 * @brief Macro defines for UART Modem Control Register
 */
#define UART_MCR_DTR_CTRL       (1 << 0)		/*!< Source for modem output pin DTR */
#define UART_MCR_RTS_CTRL       (1 << 1)		/*!< Source for modem output pin RTS */
#define UART_MCR_LOOPB_EN       (1 << 4)		/*!< Loop back mode select */
#define UART_MCR_AUTO_RTS_EN    (1 << 6)		/*!< Enable Auto RTS flow-control */
#define UART_MCR_AUTO_CTS_EN    (1 << 7)		/*!< Enable Auto CTS flow-control */
#define UART_MCR_BITMASK        (0xD3)			/*!< UART bit mask value */

/**
 * @brief Macro defines for UART Line Status Register
 */
#define UART_LSR_RDR        (1 << 0)	/*!< Line status: Receive data ready */
#define UART_LSR_OE         (1 << 1)	/*!< Line status: Overrun error */
#define UART_LSR_PE         (1 << 2)	/*!< Line status: Parity error */
#define UART_LSR_FE         (1 << 3)	/*!< Line status: Framing error */
#define UART_LSR_BI         (1 << 4)	/*!< Line status: Break interrupt */
#define UART_LSR_THRE       (1 << 5)	/*!< Line status: Transmit holding register empty */
#define UART_LSR_TEMT       (1 << 6)	/*!< Line status: Transmitter empty */
#define UART_LSR_RXFE       (1 << 7)	/*!< Line status: Error in RX FIFO */
#define UART_LSR_TXFE       (1 << 8)	/*!< Line status: Error in RX FIFO */
#define UART_LSR_BITMASK    (0xFF)		/*!< UART Line status bit mask */
#define UART1_LSR_BITMASK   (0x1FF)		/*!< UART1 Line status bit mask - valid for 11xx, 18xx/43xx UART0/2/3 only */

/**
 * @brief Macro defines for UART Modem Status Register
 */
#define UART_MSR_DELTA_CTS      (1 << 0)	/*!< Modem status: State change of input CTS */
#define UART_MSR_DELTA_DSR      (1 << 1)	/*!< Modem status: State change of input DSR */
#define UART_MSR_LO2HI_RI       (1 << 2)	/*!< Modem status: Low to high transition of input RI */
#define UART_MSR_DELTA_DCD      (1 << 3)	/*!< Modem status: State change of input DCD */
#define UART_MSR_CTS            (1 << 4)	/*!< Modem status: Clear To Send State */
#define UART_MSR_DSR            (1 << 5)	/*!< Modem status: Data Set Ready State */
#define UART_MSR_RI             (1 << 6)	/*!< Modem status: Ring Indicator State */
#define UART_MSR_DCD            (1 << 7)	/*!< Modem status: Data Carrier Detect State */
#define UART_MSR_BITMASK        (0xFF)		/*!< Modem status: MSR register bit-mask value */

/**
 * @brief Macro defines for UART Auto baudrate control register
 */
#define UART_ACR_START              (1 << 0)	/*!< UART Auto-baud start */
#define UART_ACR_MODE               (1 << 1)	/*!< UART Auto baudrate Mode 1 */
#define UART_ACR_AUTO_RESTART       (1 << 2)	/*!< UART Auto baudrate restart */
#define UART_ACR_ABEOINT_CLR        (1 << 8)	/*!< UART End of auto-baud interrupt clear */
#define UART_ACR_ABTOINT_CLR        (1 << 9)	/*!< UART Auto-baud time-out interrupt clear */
#define UART_ACR_BITMASK            (0x307)		/*!< UART Auto Baudrate register bit mask */

/**
 * Autobaud modes
 */
#define UART_ACR_MODE0              (0)	/*!< Auto baudrate Mode 0 */
#define UART_ACR_MODE1              (1)	/*!< Auto baudrate Mode 1 */

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static TAL_COM_DCB *DCBArray[TAL_COM_PORT_MAX];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  ReadData                                                             */
/*                                                                       */
/*  Read data from the Receive Buffer Register.                          */
/*                                                                       */
/*  In    : pDCB, pUARTx                                                 */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void ReadData (TAL_COM_DCB *pDCB, UART_TypeDef *pUARTx)
{
   TAL_RESULT  Error;
   uint8_t    bData;

   /* Read data */
   bData = (uint8_t)pUARTx->RBR;
   
   /* If we have no overflow... */
   if (TAL_FALSE == pDCB->bRxOverflow)
   {
      /* ... put it into the ring buffer */
      Error = tal_MISCRingAdd(&pDCB->RxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Signal counting semaphore */
         OS_SemaSignalFromInt(&pDCB->RxRdySema);
      }
      else
      {
         /* Ups, overflow */
         pDCB->bRxOverflow = TAL_OK;
      }
   }  

} /* ReadData */

/*************************************************************************/
/*  IRQHandler                                                           */
/*                                                                       */
/*  This is the generic IRQ handler.                                     */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void IRQHandler (TAL_COM_DCB *pDCB)
{
   TAL_RESULT     Error;
   UART_TypeDef *pUARTx;   
   uint32_t      dIIR;
   uint32_t      dIER;
   uint32_t      dLSR;
   uint8_t       bIndex;
   uint8_t       bData;

   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;

   /* Get IRQ status */   
   dIIR = pUARTx->IIR;
   dLSR = pUARTx->LSR;
   dIER = pUARTx->IER;
   
   /* 
    * Check for RX interrupt
    */
    
   /* Check if FIFO is used */ 
   if (pDCB->HW.bFifoTriggerLevel != 0)
   {
      /*
       * FIFO is used
       */
       
      /* Check for a block of data, FIFO trigger level */
      if      (UART_IIR_INTID_RDA == (dIIR & UART_IIR_INTID_MASK))
      {
         /* A block of data, FIFO trigger level,  can be read */
         for (bIndex = 0; bIndex < pDCB->HW.bFifoTriggerLevel; bIndex++)
         {
            ReadData(pDCB, pUARTx);
         }
      }
      /* Check Character Time-out Indicator */
      else if (UART_IIR_INTID_CTI == (dIIR & UART_IIR_INTID_MASK))
      {
         while (dLSR & UART_LSR_RDR)
         {
            ReadData(pDCB, pUARTx);
            dLSR = pUARTx->LSR;
         }
      }
   }
   else
   {
      /*
       * FIFO is NOT used
       */
       
      /* Check if data is available */
      if (UART_IIR_INTID_RDA == (dIIR & UART_IIR_INTID_MASK))
      {
         ReadData(pDCB, pUARTx);
      } /* end "RX interrupt" */
   }      
   
   
   /* 
    * Check for TX interrupt, but only if enabled 
    */
   if ((dIER & UART_IER_THREINT) &&                            /* <= enabled ? */
       (UART_IIR_INTID_THRE == (dIIR & UART_IIR_INTID_MASK)))  /* <= TX interrupt ? */
   {
      /* Read Data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (Error != TAL_OK)
      {
         /* Ups, no data available, disable interrupt */
         pUARTx->IER &= ~UART_IER_THREINT;
      }
      else
      {
         /* Check if FIFO is used */ 
         if (pDCB->HW.bFifoTriggerLevel != 0)
         {
            /*
             * FIFO is used
             */
             
            /* Put first data into the FIFO */
            pUARTx->THR = bData;
            
            /* Get next data, sub one because the first data is in the FIFO */   
            bIndex = pDCB->HW.bFifoTriggerLevel - 1;
            while (bIndex != 0)
            {
               /* Get next data */
               Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
               if (Error != TAL_OK)
               {
                  /* No data available anymore in the TX ring buffer */
                  break;
               }
               else
               {
                  /* Put the data into the FIFO */
                  pUARTx->THR = bData;
               }
               bIndex--;
            }
         }
         else
         {
            /*
             * FIFO is NOT used
             */
            pUARTx->THR = bData; 
         }            
      }   
   } /* end "TX interrupt */
   
} /* IRQHandler */

/*************************************************************************/
/*  UARTx_IRQHandler                                                     */
/*                                                                       */
/*  This is the ARM7 UARTx IRQ handler.                                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
//static void UART0_IRQHandler (void)
static void UART0_IRQHandler (uint32_t icciar, void *context)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_1]);
   TAL_CPU_IRQ_EXIT();
} /* UART0_IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  cpu_COMInit                                                          */
/*                                                                       */
/*  Prepare the hardware for use by the Open function later. Set the HW  */
/*  information depending of ePort and "enable" the COM port.            */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMInit (TAL_COM_DCB *pDCB)
{
   TAL_RESULT   Error = TAL_ERR_COM_PORT_RANGE;
   TAL_COM_HW *pHW    = &pDCB->HW;
   
   switch (pDCB->ePort)
   {
      case TAL_COM_PORT_1:
      {
         Error = tal_BoardEnableCOM1();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_1] = pDCB;
         
            pHW->dBaseAddress = UART0_BASE;
            pHW->IrqNumber    = UART0_IRQn;
            pHW->bIrqPriority = UART0_PRIO;

            /* Set priority and function */
            tal_CPUIrqDisable(UART0_IRQn);
            tal_CPUIrqSetPriority(UART0_IRQn, UART0_PRIO);
            tal_CPUIrqSetFunction(UART0_IRQn, UART0_IRQHandler, NULL); 
         }
         break;
      } /* TAL_COM_PORT_1 */
      
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (pDCB->ePort) */
   
   return(Error);
} /* cpu_COMInit */

/*************************************************************************/
/*  cpu_COMOpen                                                          */
/*                                                                       */
/*  Open the COM port.                                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMOpen (TAL_COM_DCB *pDCB)
{
   TAL_RESULT      Error = TAL_ERROR;
   UART_TypeDef  *pUARTx;   
   uint32_t       dLCR   = 0;
   uint32_t       dDummy;
   uint32_t       dDivisor;
   uint32_t       dFrequency;

   /* Disable FIFO first */   
   pDCB->HW.bFifoTriggerLevel = 0;
   
   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;

   /* Check word length */
   switch (pDCB->Settings.eLength)
   {
      case TAL_COM_LENGTH_8:
      {
         dLCR |= UART_LCR_WLEN8;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_LENGTH;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eLength) */
   
   /* Check parity settings */
   switch (pDCB->Settings.eParity)
   {
      case TAL_COM_PARITY_NONE:
      {
         dLCR |= UART_LCR_PARITY_DIS;  /*lint !e845 */
         break;
      } 
      
      case TAL_COM_PARITY_EVEN:
      {
         dLCR |= UART_LCR_PARITY_EVEN;
         break;
      }
      
      case TAL_COM_PARITY_ODD:
      {
         dLCR |= UART_LCR_PARITY_ODD;  /*lint !e845 */
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_PARITY;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eParity) */
    
   /* Check stop bit settings */
   switch (pDCB->Settings.eStop)
   {
      case TAL_COM_STOP_1_0:
      {
         dLCR |= UART_LCR_SBS_1BIT;    /*lint !e845 */
         break;
      }
      
      case TAL_COM_STOP_2_0:
      {
         dLCR |= UART_LCR_SBS_2BIT;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_STOP;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eStop) */
   
   /* Check baud rate */
   if (pDCB->Settings.dBaudrate != 0)
   {
      /* Get UART frequency */
      alt_clk_freq_get(ALT_CLK_L4_SP, &dFrequency);

      /* Add half of the denominator to address rounding errors. */
      dDivisor = ((dFrequency + (8 * pDCB->Settings.dBaudrate)) / (16 * pDCB->Settings.dBaudrate));
   }
   else
   {
      Error = TAL_ERR_COM_BAUDRATE;
      goto COMOpenEnd;  /*lint !e801*/
   }


   /*
    * Reset UART
    */
    
   /* Write SRR::UR (Shadow Reset Register :: UART Reset) */ 
   pUARTx->SRR = ALT_UART_SRR_UR_SET_MSK; 
   
   /* Read the MSR to work around case:119085 */
   dDummy = pUARTx->MSR;


   /* 
    * Set UART register 
    */
    
   /* Disable all interrupts first */ 
   pUARTx->IER = 0x00;

   /* Set baudrate */
   pUARTx->LCR = UART_LCR_DLAB_EN;  /* Enable access to Divisor Latches */  
   pUARTx->DLL = (dDivisor & 0xFF);
   pUARTx->DLH = ((dDivisor >> 8) & 0xFF);
   pUARTx->LCR = 0;                 /* Disable access to Divisor Latches */

   /* Set line control register */
   pUARTx->LCR = dLCR;
   
   /* Set FIFO and Auto-baud Control */
   pUARTx->FCR = 0;

   /* Dummy read to clear the RBR register */
   while (UART_LSR_RDR & pUARTx->LSR)
   {
      dDummy = (uint8_t)pUARTx->RBR;
   }  
   (void)dDummy;
   
   /* Enable RX interrrupt */
   pUARTx->IER |= UART_IER_RBRINT;
   
   /* Enable the IRQ */
#if 0
   tal_CPUIrqEnable(pDCB->HW.IrqNumber);
#else

   // Set interrupt distributor target
   int cpu_target = 0x1; //CPU0 will handle the interrupts
   alt_int_dist_target_set(ALT_INT_INTERRUPT_UART0, cpu_target);
   
   // Set interrupt trigger type
//   alt_int_dist_trigger_set(ALT_INT_INTERRUPT_UART0, ALT_INT_TRIGGER_EDGE);
   
   // Enable interrupt at the distributor level
   alt_int_dist_enable(ALT_INT_INTERRUPT_UART0);
#endif

   Error = TAL_OK;
   
COMOpenEnd:   
   return(Error);
} /* cpu_COMOpen */

/*************************************************************************/
/*  cpu_COMClose                                                         */
/*                                                                       */
/*  Close the COM port.                                                  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMClose (TAL_COM_DCB *pDCB)
{
   TAL_RESULT     Error = TAL_OK;
   UART_TypeDef *pUARTx;

   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;
   
   /* Disable the IRQ */   
   tal_CPUIrqDisable(pDCB->HW.IrqNumber);

   /* Disable TX and RX interrupt  */
   pUARTx->IER = 0;
   
   return(Error);
} /* cpu_COMClose */

/*************************************************************************/
/*  cpu_COMStartTx                                                       */
/*                                                                       */
/*  Send the data from the ring buffer if the TX interrupt is disabled.  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMStartTx (TAL_COM_DCB *pDCB)
{
   TAL_RESULT     Error = TAL_OK;
   UART_TypeDef *pUARTx;
   uint8_t       bData;
   
   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pUARTx->IER & UART_IER_THREINT)
   {
      /* TX interrupt is enabled, do nothing */
   }
   else
   {
      /* Get data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Send data */
         pUARTx->THR = bData;
      
         /* Enable TX interrupt */
         pUARTx->IER |= UART_IER_THREINT;
      }               
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMStartTx */

/*************************************************************************/
/*  cpu_COMTxIsRunning                                                   */
/*                                                                       */
/*  Check if TX is still running.                                        */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT cpu_COMTxIsRunning (TAL_COM_DCB *pDCB)
{
   TAL_RESULT     Error = TAL_OK;
   UART_TypeDef *pUARTx;
   
   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pUARTx->IER & UART_IER_THREINT)
   {
      /* TX is still running */
      Error = TAL_OK;
   }
   else
   {
      /* TX is not running */
      Error = TAL_ERROR;
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMTxIsRunning */

/*** EOF ***/
