/*****************************************************************************
  Exception handlers and startup code for an CycloneV target.
  
  Copyright (c) 2016 by Michael Fischer (www.emb4fun.de).
  Based on the original Zynq_7000 source from Rowley, therefore:

  Copyright (c) 2014 Rowley Associates Limited.

  This file may be distributed under the terms of the License Agreement
  provided with this software.

  THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING THE
  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *****************************************************************************/

/*****************************************************************************
 *                           Preprocessor Definitions
 *                           ------------------------
 *
 * PRESERVE_VFP_REGISTERS
 *
 *   If defined the VFP registers are saved/restored on entry/exit of the irq_handler.
 *
 * __NO_RUNFAST_MODE
 *
 *   If defined do NOT turn on flush-to-zero and default NaN modes.
 *
 *****************************************************************************/
 
   /*
    * CR1 bits
    */ 
  .equ   CP15_CR_V, (1 << 13)
 
  .section .vectors, "ax"
  .code 32
  .align 0
  .global _vectors
  .global reset_handler
  
/*****************************************************************************
 *                                                                           *
 * Exception Vectors                                                         *
 *                                                                           *
 *****************************************************************************/
_vectors:
  ldr pc, [pc, #reset_handler_address - . - 8]  /* reset */
  ldr pc, [pc, #undef_handler_address - . - 8]  /* undefined instruction */
  ldr pc, [pc, #swi_handler_address - . - 8]    /* swi handler */
  ldr pc, [pc, #pabort_handler_address - . - 8] /* abort prefetch */
  ldr pc, [pc, #dabort_handler_address - . - 8] /* abort data */
  nop
  ldr pc, [pc, #irq_handler_address - . - 8]    /* irq */
  ldr pc, [pc, #fiq_handler_address - . - 8]    /* fiq */

reset_handler_address:
  .word reset_handler
undef_handler_address:
  .word undef_handler
swi_handler_address:
  .word swi_handler
pabort_handler_address:
  .word pabort_handler
dabort_handler_address:
  .word dabort_handler
dummy:
  .word 0   
irq_handler_address:
  .word irq_handler
fiq_handler_address:
  .word fiq_handler
  
  
  

  .section .init, "ax"
  .code 32
  .align 0

/******************************************************************************
 *                                                                            *
 * Default exception handlers                                                 *
 *                                                                            *
 ******************************************************************************/
reset_handler:

  /* Set the vector base address */
  ldr r0, =_vectors
  mcr p15, 0, r0, c12, c0, 0
  
  
  /* Check which cpu is executing (just in-case) */
  mrc   p15, 0, r0, c0, c0, 5   /* Read arm affinity register */
  ands  r0, r0, #3              /* AND with 3 since only 2 cpus available*/
  beq   primary_cpu_init

  /* If here one of secondary cpu's reached this point */
secondary_cpu_loop:
  /* wfe */
  b     secondary_cpu_loop

  /* 
   * Primary cpu init must invalidate instruction cache left-on/used by BootROM 
   */
primary_cpu_init:

  /* Clear bit V of SCTRL Register */
  mrc   p15, 0, r0, c1, c0, 0
  bic   r0,  #CP15_CR_V
  mcr   p15, 0, r0, c1, c0, 0

  /* Set up stack pointer with symbol from linker script */
  LDR   sp, =__stack_end__
  
  mov   r0, #0

  /* Invalidate TLBs */
  mcr   p15, 0, r0, c8, c7, 0   

  /* Invalidate icache */
  mcr   p15, 0, r0, c7, c5, 0
  
  /* Invalidate branch predictor */
  mcr   p15, 0, r0, c7, c5, 6

  /* DSB instruction */
  mcr   p15, 0, r0, c7, c10, 4
  
  /* ISB instruction */
  mcr   p15, 0, r0, c7, c5, 4

  /* MMU and cache setup:  clear V, C, A, M bits */
  mrc   p15, 0, r0, c1, c0, 0
  bic   r0, r0, #0x00000007
  bic   r0, r0, #0x00002000

  /* Set A, Z, I bits */
  orr   r0, r0, #0x00000002  /* Set A, strict alignment fault checking enabled. */
  orr   r0, r0, #0x00000800  /* Set Z, Program flow prediction enabled. */
  orr   r0, r0, #0x00001000  /* Set I, Instruction caching enabled. */
  mcr   p15, 0, r0, c1, c0, 0
  nop
  nop
  nop
  nop


#if !defined(__SOFTFP__)
  // enable cp11 and cp10
  mrc p15, #0x00, r0, c1, c0, #0x02
  orr r0, r0, #0x00F00000
  mcr p15, #0x00, r0, c1, c0, #0x02
  // enable VFP
  mov r0, #0x40000000
  fmxr fpexc, r0
#ifndef __NO_RUNFAST_MODE
  nop
  nop
  nop  
  vmrs r0, fpscr
  orrs r0, r0, #(0x3 << 24) // FZ and DN
  vmsr fpscr, r0
#endif
#endif

  b _start

  
/******************************************************************************
 *                                                                            *
 * Default exception handlers                                                 *
 * These are declared weak symbols so they can be redefined in user code.     * 
 *                                                                            *
 ******************************************************************************/

undef_handler:
  b .  /* Endless loop */

swi_handler:
  b .  /* Endless loop */

pabort_handler:
  b .  /* Endless loop */

dabort_handler:
  b .  /* Endless loop */



  .global irq_handler
  .type irq_handler, function
irq_handler:
  // Store the APCS registers
  sub lr, lr, #4
  push {r0-r3, r12, lr}
  
#ifdef PRESERVE_VFP_REGISTERS
  vpush {d0-d7}
  vpush {d16-d31}
  vmrs r0, fpscr
  push {r0}  
#endif

  // Ensure 8-byte stack alignment
  mov r0, sp
  and r0, r0, #4
  sub sp, sp, r0
  // Save the sp adjustment
  push {r0}

 
  // Call the ISR
  ldr r1, =alt_int_handler_irq
  blx r1

  
  // Restore the sp adjustment
  pop {r0}
  add sp, sp, r0

#ifdef PRESERVE_VFP_REGISTERS
  pop {r0}
  vmsr fpscr, r0
  vpop {d16-d31}
  vpop {d0-d7}  
#endif

  // Return from interrupt
  pop {r0-r3, r12, lr}
  movs pc, lr


fiq_handler:
  b .  /* Endless loop */

  .weak undef_handler, swi_handler, pabort_handler, dabort_handler, irq_handler, fiq_handler

/*** EOF ***/

